/* -*-  Mode:C++; c-basic-offset:4; tab-width:4; indent-tabs-mode:t -*- */
/*
 * Copyright (C) 2013-2019 by the University of Southern California
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "table.hh"
#include <iostream>
#include <sstream>
#include <cstring>
#include <cassert>
#include <err.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <algorithm>

using namespace std;

#define PIPEBUFSZ (8*1024*1024)

void check_empty(vector<string> s, const char *info)
{
	if (s.size() == 0)
		errx(1, "[error] %s is invalid", info);
}

void check_empty(string s, const char *info)
{
	if (s.length() == 0)
		errx(1, "[error] %s is invalid", info);
}

Table::Table (const string &cols, const string &fn, const vector<string> &cmp) :
	fsdb_header("#fsdb -F t " + cols + "\n"),
	compressor(cmp),
	file_name(fn)
{
	check_empty(cols, "table schema");
	check_empty(compressor, "compressor");
	check_empty(file_name, "file name");

	if (file_name != "-") {
		if ((fout = ::fopen(file_name.c_str(), "w")) == 0) {
			err(1, "[error] Unable to create file[%s]", file_name.c_str());
		}
		if (compressor[0] != "none")
			map_pipe();
	} else {
		fout = stdout;
	}

	put(fsdb_header);
}

Table::~Table()
{
	clean_up();
}

void Table::clean_up()
{
	if (fout && fout != stdout) {
		::fclose(fout);
		fout = 0;
	}
}

void Table::map_pipe()
{
	int pipefd[2];
	int i = 0, maxfd = getdtablesize();
	pid_t compressor_pid;
	if (0 != ::socketpair(AF_UNIX, SOCK_STREAM, 0, pipefd))
		err(1, "[error] fail to create pipe");
#ifdef HAVE_SETSOCKOPT
	int bufsz;
	bufsz = PIPEBUFSZ;
	if (0 != ::setsockopt(pipefd[0], SOL_SOCKET, SO_RCVBUF, &bufsz, sizeof(bufsz)))
		warn("setsockopt SO_RCVBUF failed\n");
	bufsz = PIPEBUFSZ;
	if (0 != ::setsockopt(pipefd[1], SOL_SOCKET, SO_SNDBUF, &bufsz, sizeof(bufsz)))
		warn("setsockopt SO_SNDBUF failed\n");
#endif
	compressor_pid =
# ifdef HAVE_FORK
		::fork()
# else
		::vfork()
# endif
		;

	int fd = fileno(fout);
	if (compressor_pid == -1)
		err(1, "[error] fail to fork compressor");
	if (compressor_pid == 0) { //child, aka compressor
#ifdef HAVE_SETPGID
		::setpgid(0,0);
#endif
		signal(SIGQUIT, SIG_IGN);
		signal(SIGINT, SIG_IGN);
		//need to make sure neither fd is 0 or 1 (potentially may need to swap them)
		int tmpfd = ::dup(pipefd[0]);
		if (-1 == tmpfd)
			err(1, "[error] fail to duplicate compressor tmpfd");
		//cannot be because we just called socketpair()
		assert(tmpfd != STDOUT_FILENO && tmpfd != STDIN_FILENO);

		if (-1 == ::dup2(fd, STDOUT_FILENO))
			err(1, "[error] fail to duplicate compressor stdout");

		if (-1 == ::dup2(tmpfd, STDIN_FILENO))
			err(1, "[error] fail to duplicate compressor stdin");

		for (i = 0; i<maxfd; ++i) {
			if (i != STDERR_FILENO && i != STDIN_FILENO && i != STDOUT_FILENO)
				::close(i);
		}


		const char *argv[compressor.size()+1];
		size_t i=0;
		for (string s : compressor) {
			argv[i++] = ::strdup(s.c_str());
		}
		argv[i] = NULL;

		::execvp(argv[0], const_cast<char* const*>(argv));
		err(1, "[error] failed to execute pipe command %s", argv[0]);
	} else { //parent
		close(pipefd[0]);
		dup2(pipefd[1], fd);
		close(pipefd[1]);
	}
}

void Table::put(const string &s)
{
	if (::fwrite(s.c_str(), s.length(), 1, fout) != 1)
		err(1, "[error] fail to write file");
}

void Table::finalize(const string &cmd)
{
	put(cmd);
	clean_up();
}
